@@ -162,6 +162,12 @@ public:
     uint16_t GetPort(bool byte_swap = true) const;
     bool SetPort(uint16_t port);
 
+    inline void Reset(void) {
+        addr.ss.ss_family = AF_UNSPEC;
+        cached_addr.clear();
+        cached_flags = MakeFlags::NONE;
+    }
+
     inline bool IsValid(void) const {
         return (addr.ss.ss_family != AF_UNSPEC);
     }
@@ -215,8 +221,8 @@ public:
 
     inline const std::string &
     GetString(ndFlags<MakeFlags> flags = MakeFlags::NONE) const {
-        static std::recursive_mutex lock;
-        std::lock_guard<std::recursive_mutex> lg(lock);
+        static std::mutex lock;
+        std::lock_guard<std::mutex> lg(lock);
         if (flags != cached_flags || cached_addr.empty()) {
             cached_flags = flags;
             ndAddr::MakeString(*this, cached_addr, cached_flags);
@@ -243,62 +249,9 @@ public:
         compare_flags = flags;
     }
 
-    inline bool operator==(const ndAddr &a) const {
-        if (a.addr.ss.ss_family != addr.ss.ss_family)
-            return false;
-        if (ndFlagBoolean(compare_flags, CompareFlags::PREFIX) &&
-          a.prefix != prefix)
-            return false;
-
-        switch (addr.ss.ss_family) {
-#if defined(__linux__)
-        case AF_PACKET:
-            if (! ndFlagBoolean(compare_flags, CompareFlags::ADDR))
-                return true;
-            return (memcmp(&addr.ll, &a.addr.ll,
-                      sizeof(struct sockaddr_ll)) == 0);
-#elif defined(__FreeBSD__)
-        case AF_LINK:
-            if (! ndFlagBoolean(compare_flags, CompareFlags::ADDR))
-                return true;
-            return (memcmp(&addr.dl, &a.addr.dl,
-                      sizeof(struct sockaddr_dl)) == 0);
-#endif
-        case AF_INET:
-            if (ndFlagBoolean(compare_flags, CompareFlags::ADDR) &&
-              (ndFlagBoolean(compare_flags, CompareFlags::PORT)))
-            {
-                return (memcmp(&addr.in, &a.addr.in,
-                          sizeof(struct sockaddr_in)) == 0);
-            }
-            if (ndFlagBoolean(compare_flags, CompareFlags::ADDR) &&
-              memcmp(&addr.in.sin_addr, &a.addr.in.sin_addr,
-                sizeof(struct in_addr)) == 0)
-                return true;
-            if (ndFlagBoolean(compare_flags, CompareFlags::PORT) &&
-              addr.in.sin_port == a.addr.in.sin_port)
-                return true;
-            break;
-        case AF_INET6:
-            if (ndFlagBoolean(compare_flags, CompareFlags::ADDR) &&
-              (ndFlagBoolean(compare_flags, CompareFlags::PORT)))
-            {
-                return (memcmp(&addr.in6, &a.addr.in6,
-                          sizeof(struct sockaddr_in6)) == 0);
-            }
-            if (ndFlagBoolean(compare_flags, CompareFlags::ADDR) &&
-              memcmp(&addr.in6.sin6_addr,
-                &a.addr.in6.sin6_addr,
-                sizeof(struct in6_addr)) == 0)
-                return true;
-            if (ndFlagBoolean(compare_flags, CompareFlags::PORT) &&
-              addr.in6.sin6_port == a.addr.in6.sin6_port)
-                return true;
-            break;
-        }
-        return false;
-    }
+    int Compare(const ndAddr &) const;
 
+    bool operator==(const ndAddr &a) const;
     inline bool operator!=(const ndAddr &a) const {
         return ! (a == *this);
     }
@@ -700,6 +653,33 @@ public:
         return RemoveAddress(ndAddr(addr), ifname);
     }
 
+    bool AddLinkAddress(unsigned index, const ndAddr &addr);
+    bool RemoveLinkAddress(unsigned index);
+    bool LookupLinkAddress(unsigned index, ndAddr &addr);
+
+    bool AddLinkVLANId(unsigned index, uint16_t id);
+    bool RemoveLinkVLANId(unsigned index);
+    uint16_t LookupLinkVLANId(unsigned index);
+
+    bool AddLinkName(unsigned index, const std::string &name);
+    bool RemoveLinkName(unsigned index);
+    bool LookupLinkName(unsigned index, std::string &name);
+
+    bool AddNeighbor(const ndAddr &addr, const ndAddr &lladdr);
+    bool RemoveNeighbor(const ndAddr &addr);
+    bool LookupNeighbor(const ndAddr &addr, ndAddr &lladdr);
+
+    bool AddGroupAddress(const std::string &tag, const ndAddr &addr);
+    bool RemoveGroup(const std::string &tag);
+    bool RemoveGroupAddress(const std::string &tag, const ndAddr &addr);
+    bool LookupGroupAddress(const std::string &tag, const ndAddr &addr);
+    bool GetGroupFilename(const std::string &tag, std::string &filename);
+    bool SetGroupFilename(const std::string &tag);
+    bool LoadGroupAddress(const std::string &tag, const std::string &filename);
+    bool SaveGroupAddress(const std::string &tag, const std::string &filename);
+    void LoadGroupAddresses(const std::string &path);
+    void SaveGroupAddresses(const std::string &path);
+
     void Classify(ndAddr::Type &type, const ndAddr &addr);
     inline void
     Classify(ndAddr::Type &type, const std::string &addr) {
@@ -710,15 +690,38 @@ public:
       std::set<std::string> &result, sa_family_t family = AF_UNSPEC);
 
 protected:
-    mutable std::mutex lock;
-
+    mutable std::mutex lock_ether_reserved;
     std::unordered_map<std::string, ndAddr::Type> ether_reserved;
 
+    mutable std::mutex lock_ipv4_reserved;
     nd_rn4_atype ipv4_reserved;
+    mutable std::mutex lock_ipv6_reserved;
     nd_rn6_atype ipv6_reserved;
 
-    std::unordered_map<std::string, nd_rn4_atype> ipv4_iface;
-    std::unordered_map<std::string, nd_rn6_atype> ipv6_iface;
+    mutable std::mutex lock_ipv4_ifaces;
+    std::unordered_map<std::string, nd_rn4_atype> ipv4_ifaces;
+    mutable std::mutex lock_ipv6_ifaces;
+    std::unordered_map<std::string, nd_rn6_atype> ipv6_ifaces;
+
+    mutable std::mutex lock_link_addrs;
+    std::unordered_map<unsigned, ndAddr> link_addrs;
+
+    mutable std::mutex lock_link_vlan_id;
+    std::unordered_map<unsigned, uint16_t> link_vlan_ids;
+
+    mutable std::mutex lock_link_names;
+    std::unordered_map<unsigned, std::string> link_names;
+
+    mutable std::mutex lock_neighbors;
+    typedef std::unordered_map<
+        ndAddr, ndAddr, ndAddr::ndAddrHash, ndAddr::ndAddrEqual> neighbor_map;
+    neighbor_map neighbors;
+
+    mutable std::recursive_mutex lock_groups;
+    typedef std::unordered_set<
+        ndAddr, ndAddr::ndAddrHash, ndAddr::ndAddrEqual> group_addr_set;
+    std::unordered_map<std::string, group_addr_set> groups;
+    std::unordered_map<std::string, std::string> group_filenames;
 };
 
 typedef std::unordered_set<ndAddr, ndAddr::ndAddrHash, ndAddr::ndAddrEqual> ndInterfaceAddrs;
@@ -816,12 +819,12 @@ public:
         case ndCaptureType::PCAP:
             config_pcap = iface.config_pcap;
             break;
-#if defined(_ND_ENABLE_TPACKETV3)
+#ifdef _ND_ENABLE_TPACKETV3
         case ndCaptureType::TPV3:
             config_tpv3 = iface.config_tpv3;
             break;
 #endif
-#if defined(_ND_ENABLE_NFQUEUE)
+#ifdef _ND_ENABLE_NFQUEUE
         case ndCaptureType::NFQ:
             config_nfq = iface.config_nfq;
             break;
@@ -939,15 +942,24 @@ public:
 
     inline void SetConfig(const nd_config_pcap *pcap) {
         config_pcap = *pcap;
+#ifdef _ND_ENABLE_CONNTRACK_COUNTERS
+        conntrack_counters = pcap->conntrack_counters;
+#endif
     }
-#if defined(_ND_ENABLE_TPACKETV3)
+#ifdef _ND_ENABLE_TPACKETV3
     inline void SetConfig(const nd_config_tpv3 *tpv3) {
         config_tpv3 = *tpv3;
+#ifdef _ND_ENABLE_CONNTRACK_COUNTERS
+        conntrack_counters = tpv3->conntrack_counters;
+#endif
     }
 #endif
-#if defined(_ND_ENABLE_NFQUEUE)
+#ifdef _ND_ENABLE_NFQUEUE
     inline void SetConfig(const nd_config_nfq *nfq) {
         config_nfq = *nfq;
+#ifdef _ND_ENABLE_CONNTRACK_COUNTERS
+        conntrack_counters = nfq->conntrack_counters;
+#endif
     }
 #endif
 
@@ -961,11 +973,11 @@ public:
         case ndCaptureType::PCAP:
         case ndCaptureType::PCAP_OFFLINE:
             return (config_pcap == i.config_pcap);
-#if defined(_ND_ENABLE_TPACKETV3)
+#ifdef _ND_ENABLE_TPACKETV3
         case ndCaptureType::TPV3:
             return (config_tpv3 == i.config_tpv3);
 #endif
-#if defined(_ND_ENABLE_NFQUEUE)
+#ifdef _ND_ENABLE_NFQUEUE
         case ndCaptureType::NFQ:
             return (config_nfq == i.config_nfq);
 #endif
@@ -974,18 +986,24 @@ public:
         return true;
     }
 
+    int ifindex = { -1 };
     std::string ifname;
     std::string ifname_peer;
     ndFlags<ndCaptureType> capture_type;
     ndInterfaceRole role;
 
     nd_config_pcap config_pcap;
-#if defined(_ND_ENABLE_TPACKETV3)
+#ifdef _ND_ENABLE_TPACKETV3
     nd_config_tpv3 config_tpv3;
 #endif
-#if defined(_ND_ENABLE_NFQUEUE)
+#ifdef _ND_ENABLE_NFQUEUE
     nd_config_nfq config_nfq;
 #endif
+#ifdef _ND_ENABLE_CONNTRACK_COUNTERS
+    bool conntrack_counters = { false };
+#endif
+
+    typedef std::shared_ptr<ndInterface> Ptr;
 
 protected:
     ndInterfaceAddr addrs;
@@ -993,5 +1011,3 @@ protected:
     std::atomic<bool> endpoint_snapshot;
     mutable std::mutex lock;
 };
-
-typedef std::shared_ptr<ndInterface> nd_iface_ptr;
