@@ -22,18 +22,19 @@
 
 #include <atomic>
 #include <mutex>
+#include <set>
 #include <string>
 #include <unordered_map>
 #include <vector>
 
 #include "nd-addr.hpp"
-#include "nd-apps.hpp"
 #include "nd-category.hpp"
 #include "nd-config.hpp"
 #include "nd-flags.hpp"
 #include "nd-protos.hpp"
 #include "nd-serializer.hpp"
 #include "nd-sha1.h"
+#include "nd-types.hpp"
 #include "nd-util.hpp"
 
 // TLS SNI hostname/common-name length. Reference: RFC 4366
@@ -136,7 +137,9 @@ public:
 class ndFlow : public ndSerializer
 {
 public:
-    nd_iface_ptr iface;
+    using Ptr = std::shared_ptr<ndFlow>;
+
+    ndInterface::Ptr iface;
 
     ndAddr lower_mac;
     ndAddr upper_mac;
@@ -150,6 +153,7 @@ public:
     int direction = { 0 };
 
     uint8_t ip_version = { 0 };
+    uint8_t ip_dscp = { 0 };
     uint8_t ip_protocol = { 0 };
 
     uint16_t vlan_id = { 0 };
@@ -195,9 +199,9 @@ public:
 
     ndFlags<PrivacyMask> privacy_mask = { PrivacyMask::NONE };
 
-    ndDigest digest_lower = { 0 };
+    ndDigest digest_lower = { { 0 } };
 
-    typedef std::vector<ndDigest> DigestVec;
+    using DigestVec = std::vector<ndDigest>;
     DigestVec digest_mdata;
 
     struct {
@@ -212,6 +216,7 @@ public:
         std::atomic<bool> expired = { false };
         std::atomic<bool> ip_nat = { false };
         std::atomic<bool> soft_dissector = { false };
+        std::atomic<bool> app_ip_override = { false };
     } flags;
 
     struct {
@@ -219,17 +224,28 @@ public:
         std::atomic<uint32_t> last_seq = { 0 };
     } tcp;
 
-#if defined(_ND_ENABLE_CONNTRACK) && \
-  defined(_ND_ENABLE_CONNTRACK_MDATA)
+#ifdef _ND_ENABLE_CONNTRACK
     struct {
         uint32_t id = { 0 };
+#ifdef _ND_ENABLE_CONNTRACK_MDATA
         uint32_t mark = { 0 };
+        ndAddr reply_src_addr;
+        ndAddr reply_dst_addr;
+#endif
     } conntrack;
 #endif
-
+#if defined(_ND_ENABLE_NFQUEUE)
+    struct {
+        int src_ifindex = { -1 };
+        int dst_ifindex = { -1 };
+        std::string src_iface;
+        std::string dst_iface;
+    } nfq;
+#endif
     struct {
         uint8_t version = { 0xFF };
         uint8_t ip_version = { 0 };
+        uint8_t ip_dscp = { 0 };
         uint32_t lower_teid = { 0 };
         uint32_t upper_teid = { 0 };
         ndAddr::Type lower_type = { ndAddr::Type::NONE };
@@ -245,15 +261,18 @@ public:
 
     ndProto::Id detected_protocol = { ndProto::Id::UNKNOWN };
     std::string detected_protocol_name;
+    ndpi_protocol detected_ndpi_protocol = { { 0 } };
 
-    nd_app_id_t detected_application = { ND_APP_UNKNOWN };
+    ndApp::Id detected_application = { ndApp::Id::UNKNOWN };
     std::string detected_application_name;
 
     struct {
-        nd_cat_id_t application = { ND_CAT_UNKNOWN };
-        nd_cat_id_t protocol = { ND_CAT_UNKNOWN };
-        nd_cat_id_t domain = { ND_CAT_UNKNOWN };
-        nd_cat_id_t network = { ND_CAT_UNKNOWN };
+        ndCategories::Id application = { ndCategory::UNKNOWN };
+        ndCategories::Id protocol = { ndCategory::UNKNOWN };
+        ndCategories::Id domain = { ndCategory::UNKNOWN };
+        ndCategories::Id upper_net = { ndCategory::UNKNOWN };
+        ndCategories::Id lower_net = { ndCategory::UNKNOWN };
+        ndCategories::Id overlay = { ndCategory::UNKNOWN };
     } category;
 
     struct {
@@ -278,8 +297,7 @@ public:
         std::string subject_dn;
         std::string issuer_dn;
         std::string server_cn;
-        std::string client_ja3;
-        std::string server_ja3;
+        std::string client_ja4;
         std::vector<uint8_t> cert_fingerprint;
         std::vector<std::string> alpn, alpn_server;
 
@@ -287,11 +305,6 @@ public:
             uint16_t version = { 0 };
         } ech;
 
-        struct {
-            uint16_t cipher_suite = { 0 };
-            std::string esni;
-        } esni;
-
         std::atomic<bool> proc_hello = { false };
         std::atomic<bool> proc_certificate = { false };
     } tls;
@@ -308,17 +321,24 @@ public:
         std::string domain_name;
     } mdns;
 
+    struct {
+        ndAddr mapped, peer, relayed, response, other;
+    } stun;
+
     struct {
         uint16_t ndpi_score = { 0 };
         uint16_t ndpi_score_client = { 0 };
         uint16_t ndpi_score_server = { 0 };
 
-        typedef std::set<ndRisk::Id> Risks;
+        using Risks = std::set<ndRisk::Id>;
         Risks risks;
     } risk;
 
     ndFlowStats stats;
 
+    using Tags = std::set<std::string>;
+    Tags tags;
+
     mutable std::recursive_mutex lock;
 
     uint8_t dpi_queued = { 0 };
@@ -326,14 +346,17 @@ public:
 
     struct ndpi_flow_struct *ndpi_flow = { nullptr };
 
-    ndFlow(nd_iface_ptr &iface);
+    ndFlow(ndInterface::Ptr &iface);
     ndFlow(const ndFlow &flow);
     virtual ~ndFlow();
 
     void Hash(const std::string &device, bool hash_mdata = false,
       const uint8_t *key = nullptr, size_t key_length = 0);
 
-    inline void Reset(void) { stats.Reset(); }
+    inline void Reset(bool full = false) {
+        stats.Reset();
+        if (full) vlan_id = 0;
+    }
 
     void Release(void);
 
@@ -347,15 +370,14 @@ public:
     bool HasSSHServerAgent(void) const;
     bool HasTLSClientSNI(void) const;
     bool HasTLSEncryptedCH(void) const;
-    bool HasTLSEncryptedSNI(void) const;
     bool HasTLSServerCN(void) const;
     bool HasTLSIssuerDN(void) const;
     bool HasTLSSubjectDN(void) const;
-    bool HasTLSClientJA3(void) const;
-    bool HasTLSServerJA3(void) const;
+    bool HasTLSClientJA4(void) const;
     bool HasBTInfoHash(void) const;
     bool HasSSDPUserAgent(void) const;
     bool HasMDNSDomainName(void) const;
+    bool HasSTUNAddress(void) const;
 
     enum class PrintFlags : uint8_t {
         NONE = 0,
@@ -403,6 +425,8 @@ public:
                     _upper_bytes = "other_bytes";
         std::string _lower_packets = "local_packets",
                     _upper_packets = "other_packets";
+        std::string _lower_net_cat = "local_network";
+        std::string _upper_net_cat = "other_network";
 #ifdef _ND_ENABLE_EXTENDED_STATS
         std::string _lower_rate = "local_rate";
         std::string _upper_rate = "other_rate";
@@ -431,11 +455,13 @@ public:
             _lower_port = "local_port";
             _lower_bytes = "local_bytes";
             _lower_packets = "local_packets";
+            _lower_net_cat = "local_network";
             _upper_mac = "other_mac";
             _upper_ip = "other_ip";
             _upper_port = "other_port";
             _upper_bytes = "other_bytes";
             _upper_packets = "other_packets";
+            _upper_net_cat = "other_network";
 #ifdef _ND_ENABLE_EXTENDED_STATS
             _lower_rate = "local_rate";
             _upper_rate = "other_rate";
@@ -447,11 +473,13 @@ public:
             _lower_port = "other_port";
             _lower_bytes = "other_bytes";
             _lower_packets = "other_packets";
+            _lower_net_cat = "other_network";
             _upper_mac = "local_mac";
             _upper_ip = "local_ip";
             _upper_port = "local_port";
             _upper_bytes = "local_bytes";
             _upper_packets = "local_packets";
+            _upper_net_cat = "local_network";
 #ifdef _ND_ENABLE_EXTENDED_STATS
             _lower_rate = "other_rate";
             _upper_rate = "local_rate";
@@ -492,12 +520,10 @@ public:
             }
             serialize(output, { "soft_dissector" },
               (bool)flags.soft_dissector.load());
-#if defined(_ND_ENABLE_CONNTRACK) && \
-  defined(_ND_ENABLE_CONNTRACK_MDATA)
-            serialize(output, { "ct_id" }, conntrack.id);
-            serialize(output, { "ct_mark" }, conntrack.mark);
-#endif
+            serialize(output, { "app_ip_override" },
+              (bool)flags.app_ip_override.load());
             serialize(output, { "ip_version" }, (unsigned)ip_version);
+            serialize(output, { "ip_dscp" }, (unsigned)ip_dscp);
             serialize(output, { "ip_protocol" },
               (unsigned)ip_protocol);
             serialize(output, { "vlan_id" }, (unsigned)vlan_id);
@@ -586,9 +612,39 @@ public:
               category.protocol);
             serialize(output, { "category", "domain" },
               category.domain);
-            serialize(output, { "category", "network" },
-              category.network);
+            serialize(output, { "category", _lower_net_cat },
+              category.lower_net);
+            serialize(output, { "category", _upper_net_cat },
+              category.upper_net);
+            serialize(output, { "category", "overlay" },
+              category.overlay);
 
+#ifdef _ND_ENABLE_CONNTRACK
+            serialize(output, { "conntrack", "id" }, conntrack.id);
+#ifdef _ND_ENABLE_CONNTRACK_MDATA
+            serialize(output, { "conntrack", "mark" }, conntrack.mark);
+            if (conntrack.reply_src_addr.IsValid()) {
+                serialize(output, { "conntrack", "reply_src_ip" },
+                  conntrack.reply_src_addr.GetString());
+                serialize(output, { "conntrack", "reply_src_port" },
+                  (unsigned)conntrack.reply_src_addr.GetPort());
+            }
+            if (conntrack.reply_dst_addr.IsValid()) {
+                serialize(output, { "conntrack", "reply_dst_ip" },
+                  conntrack.reply_dst_addr.GetString());
+                serialize(output, { "conntrack", "reply_dst_port" },
+                  (unsigned)conntrack.reply_dst_addr.GetPort());
+            }
+#endif // _ND_ENABLE_CONNTRACK_MDATA
+#endif // _ND_ENABLE_CONNTRACK
+#if defined(_ND_ENABLE_NFQUEUE)
+            if (! nfq.src_iface.empty()) {
+                serialize(output, { "nfq", "src_iface" }, nfq.src_iface);
+            }
+            if (! nfq.dst_iface.empty()) {
+                serialize(output, { "nfq", "dst_iface" }, nfq.dst_iface);
+            }
+#endif // _ND_ENABLE_NFQUEUE
             if (! dns_host_name.empty())
                 serialize(output, { "dns_host_name" }, dns_host_name);
 
@@ -647,16 +703,6 @@ public:
                       tohex);
                 }
 
-                if (HasTLSEncryptedSNI()) {
-                    serialize(output,
-                      { "ssl", "encrypted_sni" },
-                      tls.esni.esni);
-                    sprintf(tohex, "0x%04hx", tls.esni.cipher_suite);
-                    serialize(output,
-                      { "ssl", "esni_cipher_suite" },
-                      tohex);
-                }
-
                 if (HasTLSServerCN())
                     serialize(output,
                       { "ssl", "server_cn" }, tls.server_cn);
@@ -669,13 +715,9 @@ public:
                     serialize(output,
                       { "ssl", "subject_dn" }, tls.subject_dn);
 
-                if (HasTLSClientJA3())
+                if (HasTLSClientJA4())
                     serialize(output,
-                      { "ssl", "client_ja3" }, tls.client_ja3);
-
-                if (HasTLSServerJA3())
-                    serialize(output,
-                      { "ssl", "server_ja3" }, tls.server_ja3);
+                      { "ssl", "client_ja4" }, tls.client_ja4);
 
                 if (! tls.cert_fingerprint.empty()) {
                     nd_sha1_to_string(tls.cert_fingerprint, digest);
@@ -707,6 +749,29 @@ public:
                   mdns.domain_name);
             }
 
+            if (HasSTUNAddress()) {
+                if (stun.mapped.IsValid()) {
+                    serialize(output, { "stun", "mapped" },
+                      stun.mapped.GetString(ndAddr::MakeFlags::PORT));
+                }
+                if (stun.peer.IsValid()) {
+                    serialize(output, { "stun", "peer" },
+                      stun.peer.GetString(ndAddr::MakeFlags::PORT));
+                }
+                if (stun.relayed.IsValid()) {
+                    serialize(output, { "stun", "relayed" },
+                      stun.relayed.GetString(ndAddr::MakeFlags::PORT));
+                }
+                if (stun.response.IsValid()) {
+                    serialize(output, { "stun", "response" },
+                      stun.response.GetString(ndAddr::MakeFlags::PORT));
+                }
+                if (stun.other.IsValid()) {
+                    serialize(output, { "stun", "other" },
+                      stun.other.GetString(ndAddr::MakeFlags::PORT));
+                }
+            }
+
             serialize(output, { "first_seen_at" }, ts_first_seen);
 
             serialize(output, { "risks", "risks" }, risk.risks);
@@ -718,6 +783,8 @@ public:
             serialize(output,
               { "risks", "ndpi_risk_score_server" },
               risk.ndpi_score_server);
+
+            serialize(output, { "tags" }, tags);
         }
 
         if (ndFlagBoolean(encode_flags, EncodeFlags::TUNNELS))
@@ -764,6 +831,7 @@ public:
                 serialize(output, { "gtp", "version" }, gtp.version);
                 serialize(output, { "gtp", "ip_version" },
                   gtp.ip_version);
+                serialize(output, { "gtp", "ip_dscp" }, gtp.ip_dscp);
                 serialize(output, { "gtp", _lower_ip },
                   gtp.lower_addr.GetString());
                 serialize(output, { "gtp", _upper_ip },
@@ -818,10 +886,6 @@ public:
 #endif
         }
     }
-};
 
-typedef std::shared_ptr<ndFlow> nd_flow_ptr;
-typedef std::unordered_map<std::string, nd_flow_ptr> nd_flow_map;
-typedef std::map<std::string, nd_flow_map *> nd_flows;
-typedef std::pair<std::string, nd_flow_ptr> nd_flow_pair;
-typedef std::pair<nd_flow_map::iterator, bool> nd_flow_insert;
+    bool PushMetadataDigest(const ndDigest &digest);
+};
