@@ -27,6 +27,7 @@
 #include "nd-config.hpp"
 #include "nd-except.hpp"
 #include "nd-napi.hpp"
+#include "nd-napi-curl.hpp"
 #include "nd-sha1.h"
 #include "nd-util.hpp"
 
@@ -129,10 +130,7 @@ static int ndNetifyApiThread_progress(void *user,
     return 0;
 }
 
-ndNetifyApiThread::ndNetifyApiThread()
-  : ndThread("netify-api"), ch(nullptr), curl_rc(CURLE_OK),
-    http_rc(-1), headers_tx(nullptr) {
-
+ndNetifyApiCurl::ndNetifyApiCurl() {
     if ((ch = curl_easy_init()) == nullptr) {
         throw ndException(__PRETTY_FUNCTION__,
           "curl_easy_init");
@@ -143,47 +141,58 @@ ndNetifyApiThread::ndNetifyApiThread()
     curl_easy_setopt(ch, CURLOPT_CONNECTTIMEOUT, 20);
     curl_easy_setopt(ch, CURLOPT_TIMEOUT, 60);
     curl_easy_setopt(ch, CURLOPT_NOSIGNAL, 1);
+    curl_easy_setopt(ch, CURLOPT_SSL_VERIFYPEER, ndGC.napi_tls_verify);
+    curl_easy_setopt(ch, CURLOPT_SSL_VERIFYHOST, (ndGC.napi_tls_verify) ? 2 : 0);
+    // curl_easy_setopt(ch, CURLOPT_SSLVERSION,
+    // CURL_SSLVERSION_TLSv1);
 
-    curl_easy_setopt(ch, CURLOPT_WRITEFUNCTION,
-      ndNetifyApiThread_read_data);
-    curl_easy_setopt(ch, CURLOPT_WRITEDATA,
-      static_cast<void *>(this));
-
-    curl_easy_setopt(ch, CURLOPT_HEADERFUNCTION,
-      ndNetifyApiThread_parse_header);
-    curl_easy_setopt(ch, CURLOPT_HEADERDATA,
-      static_cast<void *>(this));
-
-    curl_easy_setopt(ch, CURLOPT_NOPROGRESS, 0);
-#if (LIBCURL_VERSION_NUM < 0x073200)
-    curl_easy_setopt(ch, CURLOPT_PROGRESSFUNCTION,
-      ndNetifyApiThread_progress);
-    curl_easy_setopt(ch, CURLOPT_PROGRESSDATA,
-      static_cast<void *>(this));
-#else
-    curl_easy_setopt(ch, CURLOPT_XFERINFOFUNCTION,
-      ndNetifyApiThread_progress);
-    curl_easy_setopt(ch, CURLOPT_XFERINFODATA,
-      static_cast<void *>(this));
-#endif
 #ifdef _ND_WITH_LIBCURL_ZLIB
 #if (LIBCURL_VERSION_NUM >= 0x072106)
     curl_easy_setopt(ch, CURLOPT_ACCEPT_ENCODING, "gzip");
 #endif
 #endif  // _ND_WITH_LIBCURL_ZLIB
+}
+
+ndNetifyApiCurl::~ndNetifyApiCurl() {
+    if (ch != nullptr) {
+        curl_easy_cleanup(ch);
+        ch = nullptr;
+    }
+}
+
+ndNetifyApiThread::ndNetifyApiThread()
+  : ndThread("netify-api") {
+    curl_private = static_cast<void *>(new ndNetifyApiCurl);
+
+    curl_easy_setopt(ndCurl->ch, CURLOPT_WRITEFUNCTION,
+      ndNetifyApiThread_read_data);
+    curl_easy_setopt(ndCurl->ch, CURLOPT_WRITEDATA,
+      static_cast<void *>(this));
+
+    curl_easy_setopt(ndCurl->ch, CURLOPT_HEADERFUNCTION,
+      ndNetifyApiThread_parse_header);
+    curl_easy_setopt(ndCurl->ch, CURLOPT_HEADERDATA,
+      static_cast<void *>(this));
+
+    curl_easy_setopt(ndCurl->ch, CURLOPT_NOPROGRESS, 0);
+#if (LIBCURL_VERSION_NUM < 0x073200)
+    curl_easy_setopt(ndCurl->ch, CURLOPT_PROGRESSFUNCTION,
+      ndNetifyApiThread_progress);
+    curl_easy_setopt(ndCurl->ch, CURLOPT_PROGRESSDATA,
+      static_cast<void *>(this));
+#else
+    curl_easy_setopt(ndCurl->ch, CURLOPT_XFERINFOFUNCTION,
+      ndNetifyApiThread_progress);
+    curl_easy_setopt(ndCurl->ch, CURLOPT_XFERINFODATA,
+      static_cast<void *>(this));
+#endif
     if (ndGC_DEBUG_CURL) {
-        curl_easy_setopt(ch, CURLOPT_VERBOSE, 1);
-        curl_easy_setopt(ch, CURLOPT_DEBUGFUNCTION,
+        curl_easy_setopt(ndCurl->ch, CURLOPT_VERBOSE, 1);
+        curl_easy_setopt(ndCurl->ch, CURLOPT_DEBUGFUNCTION,
           ndNetifyApiThread_curl_debug);
-        curl_easy_setopt(ch, CURLOPT_DEBUGDATA,
+        curl_easy_setopt(ndCurl->ch, CURLOPT_DEBUGDATA,
           static_cast<void *>(this));
     }
-
-    curl_easy_setopt(ch, CURLOPT_SSL_VERIFYPEER, ndGC.napi_tls_verify);
-    curl_easy_setopt(ch, CURLOPT_SSL_VERIFYHOST, ndGC.napi_tls_verify);
-
-    //    curl_easy_setopt(ch, CURLOPT_SSLVERSION,
-    //    CURL_SSLVERSION_TLSv1);
 }
 
 ndNetifyApiThread::~ndNetifyApiThread() {
@@ -191,12 +200,12 @@ ndNetifyApiThread::~ndNetifyApiThread() {
     Terminate();
     Join();
 
-    if (ch != nullptr) {
-        curl_easy_cleanup(ch);
-        ch = nullptr;
-    }
-
     DestroyHeaders();
+
+    if (curl_private != nullptr) {
+        delete ndCurl;
+        curl_private = nullptr;
+    }
 }
 
 void ndNetifyApiThread::AppendContent(const char *data,
@@ -246,59 +255,59 @@ void ndNetifyApiThread::CreateHeaders(const Headers &headers) {
 
     string header("User-Agent: ");
     header.append(nd_get_version_and_features());
-    headers_tx = curl_slist_append(headers_tx, header.c_str());
+    ndCurl->headers_tx = curl_slist_append(ndCurl->headers_tx, header.c_str());
 
     header = "X-Vendor-ID: ";
     header.append(ndGC.napi_vendor.c_str());
-    headers_tx = curl_slist_append(headers_tx, header.c_str());
+    ndCurl->headers_tx = curl_slist_append(ndCurl->headers_tx, header.c_str());
 
     for (auto &h : headers) {
         header = h.first;
         header.append(": ");
         header.append(h.second);
-        headers_tx = curl_slist_append(headers_tx, header.c_str());
+        ndCurl->headers_tx = curl_slist_append(ndCurl->headers_tx, header.c_str());
     }
 
-    curl_easy_setopt(ch, CURLOPT_HTTPHEADER, headers_tx);
+    curl_easy_setopt(ndCurl->ch, CURLOPT_HTTPHEADER, ndCurl->headers_tx);
 }
 
 void ndNetifyApiThread::DestroyHeaders(void) {
-    if (headers_tx != nullptr) {
-        curl_slist_free_all(headers_tx);
-        headers_tx = nullptr;
+    if (ndCurl->headers_tx != nullptr) {
+        curl_slist_free_all(ndCurl->headers_tx);
+        ndCurl->headers_tx = nullptr;
     }
 }
 
 void ndNetifyApiThread::Perform(Method method, const string &url,
   const Headers &headers, const string &payload) {
-    curl_rc = CURLE_OK;
-    curl_easy_setopt(ch, CURLOPT_URL, url.c_str());
+    ndCurl->curl_rc = CURLE_OK;
+    curl_easy_setopt(ndCurl->ch, CURLOPT_URL, url.c_str());
 
-    http_rc = -1;
+    ndCurl->http_rc = -1;
     content.clear();
     headers_rx.clear();
 
     CreateHeaders(headers);
 
-    curl_easy_setopt(ch, CURLOPT_NOBODY, 0);
+    curl_easy_setopt(ndCurl->ch, CURLOPT_NOBODY, 0);
 
     switch (method) {
     case Method::GET:
-        curl_easy_setopt(ch, CURLOPT_POST, 0);
+        curl_easy_setopt(ndCurl->ch, CURLOPT_POST, 0);
         nd_dprintf("%s: %s: %s\n", tag.c_str(), "GET", url.c_str());
         break;
     case Method::HEAD:
-        curl_easy_setopt(ch, CURLOPT_POST, 0);
-        curl_easy_setopt(ch, CURLOPT_NOBODY, 1);
+        curl_easy_setopt(ndCurl->ch, CURLOPT_POST, 0);
+        curl_easy_setopt(ndCurl->ch, CURLOPT_NOBODY, 1);
 
         nd_dprintf("%s: %s: %s\n", tag.c_str(), "HEAD",
           url.c_str());
         break;
     case Method::POST:
-        curl_easy_setopt(ch, CURLOPT_POST, 1);
-        curl_easy_setopt(ch, CURLOPT_POSTFIELDSIZE, payload.size());
+        curl_easy_setopt(ndCurl->ch, CURLOPT_POST, 1);
+        curl_easy_setopt(ndCurl->ch, CURLOPT_POSTFIELDSIZE, payload.size());
         if (payload.size()) {
-            curl_easy_setopt(ch, CURLOPT_POSTFIELDS,
+            curl_easy_setopt(ndCurl->ch, CURLOPT_POSTFIELDS,
               payload.c_str());
         }
 
@@ -307,15 +316,15 @@ void ndNetifyApiThread::Perform(Method method, const string &url,
         break;
     }
 
-    if ((curl_rc = curl_easy_perform(ch)) != CURLE_OK)
-        throw curl_rc;
+    if ((ndCurl->curl_rc = curl_easy_perform(ndCurl->ch)) != CURLE_OK)
+        throw ndCurl->curl_rc;
 
-    if ((curl_rc = curl_easy_getinfo(ch,
-           CURLINFO_RESPONSE_CODE, &http_rc)) != CURLE_OK)
-        throw curl_rc;
+    if ((ndCurl->curl_rc = curl_easy_getinfo(ndCurl->ch,
+           CURLINFO_RESPONSE_CODE, &ndCurl->http_rc)) != CURLE_OK)
+        throw ndCurl->curl_rc;
 
     char *ct = nullptr;
-    curl_easy_getinfo(ch, CURLINFO_CONTENT_TYPE, &ct);
+    curl_easy_getinfo(ndCurl->ch, CURLINFO_CONTENT_TYPE, &ct);
     if (ct != nullptr) content_type = ct;
     else {
         auto i = headers_rx.find("content-type");
@@ -386,7 +395,7 @@ void *ndNetifyApiDownload::Entry(void) {
     try {
         Perform(ndNetifyApiThread::Method::HEAD, url, headers);
 
-        if (http_rc == 200) {
+        if (ndCurl->http_rc == 200) {
             auto hdr_sha1 = headers_rx.find("x-sha1-hash");
             if (hdr_sha1 == headers_rx.end()) {
                 nd_dprintf(
@@ -403,7 +412,7 @@ void *ndNetifyApiDownload::Entry(void) {
                 if (old_hash == hdr_sha1->second) {
                     nd_dprintf(
                       "%s: file has not changed.\n", tag.c_str());
-                    http_rc = 304;
+                    ndCurl->http_rc = 304;
                     content =
                       "{\"status_code\":304, "
                       "\"status_message\":\"Not "
@@ -411,7 +420,7 @@ void *ndNetifyApiDownload::Entry(void) {
                 }
                 else {
                     nd_dprintf(
-                      "%s: file has changed, downloadig "
+                      "%s: file has changed, downloading "
                       "update...\n",
                       tag.c_str());
                     Perform(ndNetifyApiThread::Method::GET,
@@ -434,7 +443,7 @@ ndNetifyApiManager::UpdateResult ndNetifyApiManager::Update(void) {
 
     if (request != requests.end()) {
         ndNetifyApiBootstrap *bootstrap =
-          dynamic_cast<ndNetifyApiBootstrap *>(request->second);
+          static_cast<ndNetifyApiBootstrap *>(request->second);
 
         if (bootstrap->HasTerminated()) {
             result = ProcessBootstrapRequest(bootstrap);
@@ -467,8 +476,11 @@ ndNetifyApiManager::UpdateResult ndNetifyApiManager::Update(void) {
 
     size_t downloads = 0;
     static vector<Request> types = {
-        Request::DOWNLOAD_CONFIG,
+        Request::DOWNLOAD_APPLICATIONS,
+        Request::DOWNLOAD_APPLICATIONS_DAT,
         Request::DOWNLOAD_CATEGORIES,
+        Request::DOWNLOAD_INTELLIGENCE,
+        Request::DOWNLOAD_INTELLIGENCE_DAT,
     };
 
     for (auto &type : types) {
@@ -477,7 +489,7 @@ ndNetifyApiManager::UpdateResult ndNetifyApiManager::Update(void) {
         if (request != requests.end()) {
             downloads++;
             ndNetifyApiDownload *download =
-              dynamic_cast<ndNetifyApiDownload *>(request->second);
+              static_cast<ndNetifyApiDownload *>(request->second);
 
             if (download->HasTerminated()) {
                 download_results[type] =
@@ -524,10 +536,16 @@ ndNetifyApiManager::UpdateResult ndNetifyApiManager::Update(void) {
             for (auto &url : urls) {
                 string filename;
 
-                if (url.first == Request::DOWNLOAD_CONFIG)
+                if (url.first == Request::DOWNLOAD_APPLICATIONS)
                     filename = ndGC.path_app_config;
+                else if (url.first == Request::DOWNLOAD_APPLICATIONS_DAT)
+                    filename = nd_change_ext(ndGC.path_app_config, "dat");
                 else if (url.first == Request::DOWNLOAD_CATEGORIES)
                     filename = ndGC.path_cat_config;
+                else if (url.first == Request::DOWNLOAD_INTELLIGENCE)
+                    filename = ndGC.path_intel_config;
+                else if (url.first == Request::DOWNLOAD_INTELLIGENCE_DAT)
+                    filename = nd_change_ext(ndGC.path_intel_config, "dat");
 
                 ndNetifyApiDownload *download = new ndNetifyApiDownload(
                   token, url.second, filename);
@@ -570,16 +588,16 @@ ndNetifyApiManager::ProcessBootstrapRequest(
     jstatus["bootstrap"]["code"] = -1;
     jstatus["bootstrap"]["last_update"] = time(nullptr);
 
-    if (bootstrap->curl_rc != CURLE_OK) {
+    if (ndCurlParent(bootstrap)->curl_rc != CURLE_OK) {
         jstatus["bootstrap"]["message"] = curl_easy_strerror(
-          bootstrap->curl_rc);
+          ndCurlParent(bootstrap)->curl_rc);
         return UpdateResult::ERROR;
     }
     else {
         jstatus["bootstrap"]["message"] = "Unknown result";
     }
 
-    if (bootstrap->http_rc == 0) {
+    if (ndCurlParent(bootstrap)->http_rc == 0) {
         jstatus["bootstrap"]["code"] = -1;
         jstatus["bootstrap"]["message"] = "Request failure";
         nd_printf(
@@ -645,11 +663,11 @@ ndNetifyApiManager::ProcessBootstrapRequest(
         jstatus["bootstrap"]["code"] = code;
         jstatus["bootstrap"]["message"] = message;
 
-        if (bootstrap->http_rc != 200 || code != 0) {
+        if (ndCurlParent(bootstrap)->http_rc != 200 || code != 0) {
             nd_printf(
               "netify-api: Bootstrap request failed: HTTP "
               "%ld: %s [%d]\n",
-              bootstrap->http_rc, message.c_str(), code);
+              ndCurlParent(bootstrap)->http_rc, message.c_str(), code);
             return UpdateResult::ERROR;
         }
 
@@ -693,36 +711,6 @@ ndNetifyApiManager::ProcessBootstrapRequest(
             return UpdateResult::ERROR;
         }
 
-        auto japps = jsigs->find("applications_endpoint");
-        if (japps == jsigs->end() ||
-          japps->type() != json::value_t::string)
-        {
-            jstatus["bootstrap"]["code"] = -1;
-            jstatus["bootstrap"]["message"] =
-              "Application signature endpoints not found";
-            nd_dprintf(
-              "netify-api: Malformed bootstrap content: "
-              "%s\n",
-              "applications_endpoint not found or invalid "
-              "type");
-            return UpdateResult::ERROR;
-        }
-
-        auto jcats = jsigs->find("categories_endpoint");
-        if (jcats == jsigs->end() ||
-          jcats->type() != json::value_t::string)
-        {
-            jstatus["bootstrap"]["code"] = -1;
-            jstatus["bootstrap"]["message"] =
-              "Category index endpoints not found";
-            nd_dprintf(
-              "netify-api: Malformed bootstrap content: "
-              "%s\n",
-              "categories_endpoint not found or invalid "
-              "type");
-            return UpdateResult::ERROR;
-        }
-
         auto jtoken = jsigs->find("token");
         if (jtoken == jsigs->end() ||
           jtoken->type() != json::value_t::string)
@@ -744,8 +732,35 @@ ndNetifyApiManager::ProcessBootstrapRequest(
             nd_dprintf("netify-api: new API token set.\n");
         }
 
-        urls[Request::DOWNLOAD_CONFIG] = japps->get<string>();
-        urls[Request::DOWNLOAD_CATEGORIES] = jcats->get<string>();
+        auto japps = jsigs->find("applications_endpoint");
+        if (japps != jsigs->end() ||
+          japps->type() == json::value_t::string) {
+            urls[Request::DOWNLOAD_APPLICATIONS] = japps->get<string>();
+        }
+
+        auto japps_dat = jsigs->find("applications_dat_endpoint");
+        if (japps_dat != jsigs->end() &&
+          japps_dat->type() == json::value_t::string) {
+            urls[Request::DOWNLOAD_APPLICATIONS_DAT] = japps_dat->get<string>();
+        }
+
+        auto jcats = jsigs->find("categories_endpoint");
+        if (jcats != jsigs->end() ||
+          jcats->type() == json::value_t::string) {
+            urls[Request::DOWNLOAD_CATEGORIES] = jcats->get<string>();
+        }
+
+        auto jintel = jsigs->find("intelligence_endpoint");
+        if (jintel != jsigs->end() &&
+          jintel->type() == json::value_t::string) {
+            urls[Request::DOWNLOAD_INTELLIGENCE] = jintel->get<string>();
+        }
+
+        auto jintel_dat = jsigs->find("intelligence_dat_endpoint");
+        if (jintel_dat != jsigs->end() &&
+          jintel_dat->type() == json::value_t::string) {
+            urls[Request::DOWNLOAD_INTELLIGENCE_DAT] = jintel_dat->get<string>();
+        }
     }
     catch (json::exception &e) {
         jstatus["bootstrap"]["code"] = -1;
@@ -767,27 +782,36 @@ bool ndNetifyApiManager::ProcessDownloadRequest(
     string status_type;
 
     switch (type) {
-    case Request::DOWNLOAD_CONFIG:
+    case Request::DOWNLOAD_APPLICATIONS:
         status_type = "applications";
         break;
+    case Request::DOWNLOAD_APPLICATIONS_DAT:
+        status_type = "applications (dat)";
+        break;
     case Request::DOWNLOAD_CATEGORIES:
         status_type = "categories";
         break;
+    case Request::DOWNLOAD_INTELLIGENCE:
+        status_type = "intelligence";
+        break;
+    case Request::DOWNLOAD_INTELLIGENCE_DAT:
+        status_type = "intelligence (dat)";
+        break;
     default:
         nd_dprintf("netify-api: invalid download type: %d", type);
         return false;
     }
 
-    jstatus[status_type]["code"] = download->http_rc;
+    jstatus[status_type]["code"] = ndCurlParent(download)->http_rc;
     jstatus[status_type]["last_update"] = time(nullptr);
 
-    if (download->curl_rc != CURLE_OK) {
+    if (ndCurlParent(download)->curl_rc != CURLE_OK) {
         jstatus[status_type]["message"] = curl_easy_strerror(
-          download->curl_rc);
+          ndCurlParent(download)->curl_rc);
         return false;
     }
 
-    switch (download->http_rc) {
+    switch (ndCurlParent(download)->http_rc) {
     case 200:
         jstatus[status_type]["message"] = "Updated";
         break;
@@ -808,14 +832,14 @@ bool ndNetifyApiManager::ProcessDownloadRequest(
         break;
     }
 
-    if (download->http_rc != 200) {
+    if (ndCurlParent(download)->http_rc != 200) {
         nd_printf(
           "netify-api: Download request failed: HTTP %ld: "
           "type: %d\n",
-          download->http_rc, type);
+          ndCurlParent(download)->http_rc, type);
 
-        if (download->http_rc == 401 || download->http_rc == 403)
-        {
+        if (ndCurlParent(download)->http_rc == 401 ||
+            ndCurlParent(download)->http_rc == 403) {
             nd_dprintf(
               "netify-api: cleared token on authorization "
               "failure.\n");
@@ -826,16 +850,31 @@ bool ndNetifyApiManager::ProcessDownloadRequest(
         return false;
     }
 
-    if (type == Request::DOWNLOAD_CONFIG) {
+    if (type == Request::DOWNLOAD_APPLICATIONS) {
         return nd_copy_file(download->content_filename,
           ndGC.path_app_config,
           S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
     }
+    else if (type == Request::DOWNLOAD_APPLICATIONS_DAT) {
+        return nd_copy_file(download->content_filename,
+          nd_change_ext(ndGC.path_app_config, "dat"),
+          S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
+    }
     else if (type == Request::DOWNLOAD_CATEGORIES) {
         return nd_copy_file(download->content_filename,
           ndGC.path_cat_config,
           S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
     }
+    else if (type == Request::DOWNLOAD_INTELLIGENCE) {
+        return nd_copy_file(download->content_filename,
+          ndGC.path_intel_config,
+          S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
+    }
+    else if (type == Request::DOWNLOAD_INTELLIGENCE_DAT) {
+        return nd_copy_file(download->content_filename,
+          nd_change_ext(ndGC.path_intel_config, "dat"),
+          S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
+    }
 
     return false;
 }
